/* GPS display software version 1.0 of 24 september 2015
The GPS display is an Arduino version of the original Roverbox of ON4IY.
Ideas and part of the code are taken from the internet e.g.from ON4IY, ON7EQ, K3NG etc.
comments are welcome. ON4CDU (at) uba.be
*/

#include <LiquidCrystal.h> // include the liquid crystal library code
LiquidCrystal lcd(12, 11, 10, 9, 8, 7);// initialize the library with the numbers of the interface pins

const int GPSbitrate = 9600; // bitrate GNSS receiver
#define HEADER_RMC "$GPRMC"
#define HEADER_GGA "$GPGGA"
const int pinled = 13;
const int EOled = 4;
const int sentenceSize = 82;
const int max_nr_words = 20;
char sentence[sentenceSize], sz[20];
char* words [max_nr_words];
char EO, sts, latdir, longdir;
int alternate;
int hours, minutes, seconds, year, month, day;
int gps_fixval, gps_sats, alti, hdopi;
float time, date, latf, lonf, hdopf;
boolean status, gps_fix, hdop;
char char_string [] = "ABCDEFGHIJKLMNOPQRSTUVWXYZ";
char num_string[]= "0123456789";

// NMEA fields
#define RMC_TIME 1
#define RMC_STATUS 2
#define RMC_LAT 3
#define RMC_NS 4
#define RMC_LONG 5
#define RMC_EW 6
#define RMC_DATE 9
#define GGA_FIX 6
#define GGA_SATS 7
#define GGA_HDOP 8
#define GGA_ALTITUDE 9

void setup()
{
  Serial.begin(GPSbitrate);
  pinMode (pinled, OUTPUT);
  pinMode (EOled, OUTPUT);
  lcd.begin(16, 2); // set up the LCD's number of columns and rows
  lcd.setCursor (0, 0);
  digitalWrite (EOled, LOW); // EO led off
  lcd.print ("waiting for GPS");
}

void loop()
{
  digitalWrite (pinled, LOW);
  static int i = 0;
  if (Serial.available())
  {
    digitalWrite (pinled, HIGH); 
    char ch = Serial.read();
    if (ch != '\n' && i < sentenceSize)
    {
      sentence[i] = ch;
      i++;
    }
    else
    {
      sentence[i] = '\0';
      i = 0;
      displayGPS();
    }
  }
}

void displayGPS()
{
  getField ();
  if (strcmp(words [0], HEADER_RMC) == 0)
  {
    processRMCstring();
    printfirstrow ();
    locator();
  }
  else if (strcmp(words[0], HEADER_GGA) == 0)
  {
    processGGAstring ();
  }
  printGGAinfo ();
}

void getField () // split string was probably a better name
{
  char* p_input = &sentence[0];
  words [0] = strsep (&p_input, ",");
  for (int i=1; i<max_nr_words; ++i)
  {
    words [i] = strsep (&p_input, ",");
  }
}

void processRMCstring ()
{
    time = atof(words [RMC_TIME]);
    hours = (int)(time * 0.0001);
    minutes = int((time * 0.01) - (hours * 100));
    seconds = int(time - (hours * 10000) - (minutes * 100));

    if (minutes % 2 > 0)
    {
      EO = 'O';
      digitalWrite (EOled, LOW); // EO led off
    }
    else
    {
      EO = 'E';
      digitalWrite (EOled, HIGH); // EO led on
    }

    sts = words[RMC_STATUS][0]; // status
    status = (sts == 'A');
 
    latf = atof(words [RMC_LAT]); // latitude
    latf = int (latf * 0.01) + (((latf * 0.01) - int(latf * 0.01)) / 0.6);

    latdir = words[RMC_NS][0]; // N/S

    lonf= atof(words [RMC_LONG]);// longitude
    lonf = int (lonf * 0.01) + ((lonf * 0.01 - int(lonf * 0.01)) / 0.6);

    longdir = words[RMC_EW][0]; // E/W

    date = atof(words [RMC_DATE]); // date
    day = int(date * 0.0001);
    month = int ((date * 0.01) - (day * 100));
    year = int (date - (day * 10000) - (month * 100));
}

void processGGAstring ()
{
 gps_fixval = words[GGA_FIX][0] - 48;
    gps_fix = (gps_fixval > 0);


    gps_sats = atoi(words[GGA_SATS]);

    hdopf = atof (words[GGA_HDOP]);
    if (hdopf > 1) 
    {
      hdopi = 2;
      hdop = true;
    }
    else if (hdopf > 2) {
      hdopi = 3;
      hdop = false;
    }
    else {
      hdopi = 1;
      hdop = true;
    }
    
    alti = atoi(words[GGA_ALTITUDE]); 
}

void printfirstrow ()
{
    sprintf(sz, "%02d/%02d%02c %02d:%02d:%02d ",
            day, month , EO , hours , minutes , seconds );
    lcd.setCursor(0, 0);
    lcd.print(sz); // print time, EO and date string on LCD  
}

void printGGAinfo ()
{
  if (status && gps_fix && hdop)
  {
    alternate = seconds / 20;
    
    switch (alternate)
    {
      case 0:
      {
        sprintf (sz, "! %2d SATS" , gps_sats);
        break;
      }
      case 1:
      {
        sprintf (sz, "!  HDOP %1d" , hdopi);
        break;
      }
      case 2:
      {
        sprintf (sz, "!   %3d m" , alti);
        break;
      }
    }
  }
  else
  {
    alternate = (seconds % 2);
    if (0 == alternate)
    {
      sprintf (sz, "? %2d SATS" , gps_sats);
    }
    else
    {
      sprintf (sz, "  %2d SATS" , gps_sats);
    }
  }
  
  lcd.setCursor(7, 1);
  lcd.print (sz);
}

void locator() // locator calculation and display
{
  float loclong = lonf * 1000000;
  float loclat = latf * 1000000;
  float scrap;

  if (longdir == 'E') 
  {
    loclong = (loclong) + 180000000;
  }
  if (longdir == 'W') 
  {
    loclong = 180000000 - (loclong);
  }

  if (latdir == 'N') 
  {
    loclat = loclat + 90000000;
  }
  if (latdir == 'S') 
  {
    loclat = 90000000 - loclat;
  }

  lcd.setCursor(0, 1); // printing QTH locator

  // First Character - longitude based (every 20° = 1 gridsq)
  lcd.print(char_string[int(loclong / 20000000)]);

  // Second Character - latitude based (every 10° = 1 gridsq)
  lcd.print(char_string[int(loclat / 10000000)]);

  // Third Character - longitude based (every 2° = 1 gridsq)
  scrap = loclong - (20000000 * int(loclong / 20000000));
  lcd.print(num_string[int(scrap * 10 / 20 / 1000000)]);

  // Fourth Character - latitude based (every 1° = 1 gridsq)
  scrap = loclat - (10000000 * int(loclat / 10000000));
  lcd.print(num_string[int(scrap / 1000000)]);

  // Fifth Character - longitude based (every 5' = 1 gridsq)
  scrap = (loclong / 2000000) - (int(loclong / 2000000));
  lcd.print(char_string[int(scrap * 24)]);

  // Sixth Character - longitude based (every 2.5' = 1 gridsq)
  scrap = (loclat / 1000000) - (int(loclat / 1000000));
  lcd.print(char_string[int(scrap * 24)]);
}

